from rich.console import Console, ConsoleOptions, RenderResult
from rich.markdown import Heading, Markdown


class NotebookHeading(Heading):
    "A Markdown heading inside a notebook Markdown cell."

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        text = self.text
        if self.level == 1:
            # Double underline for h1s
            yield text
            yield "=" * len(text)
        # elif self.level == 2:
        #     # Single underline for h2s
        #     yield text
        #     yield "-" * len(text)
        else:
            # No underline for h3s and beyond
            yield text


class NotebookMarkdown(Markdown):
    "A renderable for Markdown text in Markdown cells."

    elements = Markdown.elements
    elements["heading"] = NotebookHeading



"""Linting functions for notebooks."""
import ast
import re
from typing import List

from . import lint_register as register
from .cell import Cell, CellType
from .config import settings
from .lint import LintDefinition, LintLevel
from .notebook import Notebook

# ============== #
# NOTEBOOK LEVEL #
# ============== #


def non_linear_execution(notebook: Notebook) -> bool:
    """Check linear execution order of notebook cells."""
    exec_counters: List[int] = [
        cell.exec_count for cell in notebook.code_cells if cell.exec_count
    ]
    sorted_counters = sorted(exec_counters)
    return exec_counters != sorted_counters


def notebook_too_long(notebook: Notebook) -> bool:
    """Check if the notebook is too long (i.e., if it contains too many cells)."""

    return len(notebook) > settings.max_cells_in_notebook


def untitled_notebook(notebook: Notebook) -> bool:
    """Check whether the notebook is untitled.

    I.e., if it was left with the default title: ``Untitled.ipynb``.
    """
    return notebook.path.name == "Untitled.ipynb"


def notebook_named_with_unrestricted_charset(notebook: Notebook) -> bool:
    """Check if the notebook filename contains characters outside ``[A-Za-z0-9_.-]``.

    To be supported by all popular operating systems,
    notebook names should be restricted to the ``[A-Za-z0-9_.-]`` charset.
    """
    return not re.search("^[A-Za-z0-9_.-]+$", notebook.path.name)


def long_filename(notebook: Notebook) -> bool:
    """Check if the notebook title exceedes the fixed character threshold."""
    if settings.filename_max_length:
        return len(notebook.path.name) > settings.filename_max_length
    else:
        return False


def imports_beyond_first_cell(notebook: Notebook) -> bool:
    """Check if import statements are used beyond the first code cell."""

    code = notebook.script
    found_first_cell = False
    # when `found_first_cell` is True, it means we have found the first cell of code
    # that has to be ignored

    second_cell_not_reached = True
    # when set to False, we are actually reading instructions from the second cell of
    # code; from now on we need to analyze all the cells looking for import statements

    correct_position = True
    cell = ""
    program = code.split("\n")
    for line in program:
        if not found_first_cell:
            # it ignores all the lines before the first cell generated by
            # nbconvert(python# version ecc.)
            if line[0:5] == "# In[":
                found_first_cell = True
        elif not second_cell_not_reached:
            # starting from the second cell, it saves all the instructions until
            # it finds a new cell
            if line[0:5] != "# In[":
                cell = cell + "\n" + line
            else:
                tree = ast.parse(cell)
                # once it finds a new cell, it checks if there are any import statements
                # in the previous cell
                if sum(isinstance(exp, ast.Import) for exp in tree.body) > 0:
                    correct_position = False
                    break
        else:
            if line[0:5] == "# In[":
                # following instructions are from the second cell of code,
                # the first one we have to analyze
                second_cell_not_reached = False
    return not correct_position


def missing_h1_md_heading(notebook: Notebook) -> bool:
    """Check that the notebook has an H1 Markdown title in the initial cells.

    Args:
        notebook (Notebook): the notebook to be analyzed.

    Returns:
        bool: ``True`` if the notebook does not contain an H1 title in the selected set
            of initial cells; ``False`` otherwise.
    """
    md_rows = "\n".join(
        [
            cell.cell_source
            for cell in notebook.initial_cells
            if cell.cell_type == CellType.MARKDOWN
        ]
    )
    pattern = re.compile(r"^\s*#\s*[^#\n]*$")
    return not any([pattern.match(line) for line in md_rows.splitlines()])


def missing_opening_MD_text(notebook: Notebook) -> bool:
    """Check that descriptive MD cells are present among the first cells of a notebook.

    Markdown cells containing just Markdown headings do not count.

    Args:
        notebook (Notebook): the notebook to be analyzed.

    Returns:
        bool: ``True`` if the notebook has no MD cells among its fist cells;
            ``False`` otherwise.
    """
    return not any(
        cell.cell_type == CellType.MARKDOWN and not cell.is_heading
        for cell in notebook.initial_cells
    )


def missing_closing_MD_text(notebook: Notebook) -> bool:
    """Check that descriptive MD cells are present among the first cells of a notebook.

    Markdown cells containing just Markdown headings do not count.

    Args:
        notebook (Notebook): the notebook to be analyzed.

    Returns:
        bool: ``True`` if the notebook has no MD cells among its fist cells;
            ``False`` otherwise.
    """
    return not any(
        cell.cell_type == CellType.MARKDOWN and not cell.is_heading
        for cell in notebook.final_cells
    )


def too_few_MD_cells(notebook: Notebook) -> bool:
    """Check that the number of MD cells is adequate.

    Check that the number of MD cells is adequate with respect
    to the number of code cells.

    Args:
        notebook (Notebook): the notebook to be analyzed.

    Returns:
        bool: ``True`` if the notebook contains too few MD cells with respect
        to the existing code cells; ``False`` otherwise.
    """
    n_of_md_cells = len(notebook.markdown_cells)
    n_of_code_cells = len(notebook.code_cells)
    if n_of_code_cells:
        ratio = n_of_md_cells / n_of_code_cells
        return ratio < settings.min_md_code_ratio
    else:
        return False


# ========== #
# CELL LEVEL #
# ========== #


def non_executed_cells(notebook: Notebook) -> List[Cell]:
    """Check the existence of non executed cells and return their list."""
    return [cell for cell in notebook.code_cells if cell.non_executed]


def empty_cells(notebook: Notebook) -> List[Cell]:
    """Check the existence of empty cells and return their list."""
    return [cell for cell in notebook.code_cells if cell.empty]


def cells_too_long(notebook: Notebook) -> List[Cell]:
    """Check whether code cells in this notebook are too long."""
    return [
        cell
        for cell in notebook.code_cells
        if len(cell.cell_source.split("\n")) > settings.max_lines_in_code_cell
    ]


# ================= #
# LINT REGISTRATION #
# ================= #


notebook_level_lints: List[LintDefinition] = [
    LintDefinition(
        slug="non-linear-execution",
        description="Notebook cells have been executed in a non-linear order.",
        recommendation="Re-run your notebook top to bottom to ensure it is "
        "reproducible.",
        linting_function=non_linear_execution,
    ),
    LintDefinition(
        slug="notebook-too-long",
        description="The notebook is too long: the total number of cells exceeds "
        f"the fixed threshold ({settings.max_cells_in_notebook}).",
        recommendation="Split this notebook into two or more notebooks.",
        linting_function=notebook_too_long,
    ),
    LintDefinition(
        slug="untitled-notebook",
        description='The notebook still has the default title: "Untitled.ipynb".',
        recommendation="Give it a meaningful title to make it easy to recognize.",
        linting_function=untitled_notebook,
    ),
    LintDefinition(
        slug="non-portable-chars-in-nb-name",
        description="The notebook filename contains non-portable characters "
        "(i.e., characters outside the [A-Za-z0-9_.-] charset).",
        recommendation="Rename your notebook by using characters contained "
        "in the following portable charset: [A-Za-z0-9_.-].",
        linting_function=notebook_named_with_unrestricted_charset,
    ),
    LintDefinition(
        slug="notebook-name-too-long",
        description="The notebook filename is too long (i.e., it exceeds the "
        f"fixed threshold of {settings.filename_max_length} characters).",
        recommendation="Use a shorter filename and leverage Markdown titles to convey "
        "detailed information.",
        linting_function=long_filename,
    ),
    LintDefinition(
        slug="imports-beyond-first-cell",
        description="Import statements found beyond the first cell of the notebook.",
        recommendation="Move import statements to the first code cell to make "
        "your notebook dependencies more explicit.",
        linting_function=imports_beyond_first_cell,
    ),
    LintDefinition(
        slug="missing-h1-MD-heading",
        description="An H1 Markdown heading is missing from the initial cells "
        "of the notebook.",
        recommendation="Clarify the notebook subject by writing an H1 Markdown heading "
        "in one of the initial cells of your notebook.",
        linting_function=missing_h1_md_heading,
    ),
    LintDefinition(
        slug="missing-opening-MD-text",
        description="The initial notebook cells "
        f"(i.e., the first {settings.initial_cells} cells in the notebook) "
        "contain no Markdown text.",
        recommendation="Begin your notebook by describing what you intend to do "
        "in one or more introductory Markdown cells.",
        linting_function=missing_opening_MD_text,
    ),
    LintDefinition(
        slug="missing-closing-MD-text",
        description="The final notebook cells "
        f"(i.e., the last {settings.final_cells} cells in the notebook) "
        "contain no Markdown text.",
        recommendation="Conclude your notebook by describing what you have accomplished"
        " in one or more concluding Markdown cells.",
        linting_function=missing_closing_MD_text,
    ),
    LintDefinition(
        slug="too-few-MD-cells",
        description="The notebook contains too few Markdown cells compared to code "
        "cells (the ratio is below the fixed threshold of "
        f"{settings.min_md_code_ratio*100}%).",
        recommendation="Describe the steps of your computation by adding "
        "a few more Markdown cells.",
        linting_function=too_few_MD_cells,
    ),
]

cell_level_lints: List[LintDefinition] = [
    LintDefinition(
        slug="non-executed-cells",
        description="Non-executed cells are present in the notebook.",
        recommendation="Re-run your notebook top to bottom to ensure that all cells "
        "are executed.",
        linting_function=non_executed_cells,
    ),
    LintDefinition(
        slug="empty-cells",
        description="Empty cells are present in the notebook.",
        recommendation="Keep your notebook clean by deleting unused cells.",
        linting_function=empty_cells,
        show_details=False,
    ),
    LintDefinition(
        slug="cell-too-long",
        description="One or more code cells in this notebook are too long "
        "(i.e., they exceed the fixed threshold "
        f"of {settings.max_lines_in_code_cell} lines).",
        recommendation="Consider consolidating your code outside the notebook "
        "by moving utility functions to a structured and tested codebase.\n"
        "Use notebooks to display results, not to compute them.",
        linting_function=cells_too_long,
    ),
]


def initialize() -> None:
    register.register_lints(LintLevel.NOTEBOOK, notebook_level_lints)
    register.register_lints(LintLevel.CELL, cell_level_lints)



import ast
import dataclasses
from dataclasses import dataclass
from typing import Dict, List, Optional

from pydantic import BaseModel
from rich.columns import Columns
from rich.console import Console, ConsoleOptions, RenderResult, group
from rich.panel import Panel
from rich.rule import Rule

from .lint import CellLevelLint, NotebookLevelLint, NotebookLint
from .lint_register import enabled_cell_level_lints, enabled_notebook_level_lints
from .notebook import Notebook


class NotebookLinterOptions(BaseModel):
    bottom_size: int = 4
    filename_max_length: Optional[int] = None


@dataclass
class NotebookMetadata:
    notebook_name: str


@dataclass
class NotebookStats:
    # Cells
    number_of_cells: int
    number_of_MD_cells: int
    number_of_code_cells: int
    number_of_raw_cells: int

    # Modularization
    number_of_functions: int
    number_of_classes: int

    # Markdown usage
    number_of_md_lines: int
    number_of_md_titles: int


class NotebookLinter:
    def __init__(self, notebook: Notebook) -> None:
        self.notebook = notebook
        self.options: NotebookLinterOptions = NotebookLinterOptions()
        self.notebook_metadata: NotebookMetadata = NotebookMetadata(
            notebook_name=notebook.path.name
        )
        self.notebook_stats: NotebookStats = NotebookStats(
            number_of_cells=self.count_cells(),
            number_of_MD_cells=self.count_md_cells(),
            number_of_code_cells=self.count_code_cells(),
            number_of_raw_cells=self.count_raw_cells(),
            number_of_functions=self.count_func_defs(),
            number_of_classes=self.count_class_defs(),
            number_of_md_lines=self.count_md_lines(),
            number_of_md_titles=self.count_md_titles(),
        )

        self.lints: List[NotebookLint] = []

        self.lints.extend(
            [
                NotebookLevelLint(
                    lint.slug,
                    lint.description,
                    lint.recommendation,
                    lint.linting_function,
                    self.notebook,
                )
                for lint in enabled_notebook_level_lints
            ]
        )

        self.lints.extend(
            [
                CellLevelLint(
                    lint.slug,
                    lint.description,
                    lint.recommendation,
                    lint.linting_function,
                    self.notebook,
                    lint.show_details,
                )
                for lint in enabled_cell_level_lints
            ]
        )

        self.has_linting_results = any([lint.result for lint in self.lints])

    def count_cells(self) -> int:
        """Computes the total number of cells within a notebook."""

        nb_dict = self.notebook.nb_dict
        return len(nb_dict["cells"])

    def count_md_cells(self) -> int:
        """Computes the total number of Markdown cells within a notebook."""

        nb_dict = self.notebook.nb_dict
        counter = 0
        for cell in nb_dict["cells"]:
            if cell["cell_type"] == "markdown":
                counter = counter + 1
        return counter

    def count_code_cells(self) -> int:
        """Computes the total number of code cells within a notebook."""

        nb_dict = self.notebook.nb_dict
        counter = 0
        for cell in nb_dict["cells"]:
            if cell["cell_type"] == "code":
                counter = counter + 1
        return counter

    def count_raw_cells(self) -> int:
        """Computes the total number of raw cells within a notebook."""

        nb_dict = self.notebook.nb_dict
        counter = 0
        for cell in nb_dict["cells"]:
            if cell["cell_type"] == "raw":
                counter = counter + 1
        return counter

    def count_func_defs(self):
        """Computes the total number of function definitions within a notebook."""
        code = self.notebook.script
        tree = ast.parse(code)
        f_num = sum(isinstance(exp, ast.FunctionDef) for exp in tree.body)
        return f_num

    def count_class_defs(self):
        """Computes the total number of class definitions within a notebook."""
        code = self.notebook.script
        tree = ast.parse(code)
        class_def_num = sum(isinstance(exp, ast.ClassDef) for exp in tree.body)
        return class_def_num

    def count_md_lines(self) -> int:
        """Count the total number of markdown rows within a notebook."""
        nb_dict = self.notebook.nb_dict
        markdown_lines = 0
        for cell in nb_dict["cells"]:
            if cell["cell_type"] == "markdown":
                rows = len(cell["source"].split("\n"))
                markdown_lines += rows
        return markdown_lines

    def count_md_titles(self) -> int:
        """Count the total number of markdown titles within a notebook."""
        nb_dict = self.notebook.nb_dict
        titles = 0
        for cell in nb_dict["cells"]:
            if cell["cell_type"] == "markdown":
                for row in cell["source"]:
                    if row.lstrip().startswith("#"):
                        titles = titles + 1
        return titles

    def as_dict(self) -> Dict:
        results_dict = {
            "notebook_metadata": dataclasses.asdict(self.notebook_metadata),
            "notebook_stats": dataclasses.asdict(self.notebook_stats),
            "lints": [lint.as_dict() for lint in self.lints if lint.result],
        }
        return results_dict

    @group()
    def get_renderable_linting_results(self):
        for lint in self.lints:
            if lint.result:
                yield lint

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        # Notebook name and path
        notebook_name = "\n"
        notebook_name += "[blue bold underline]NOTEBOOK[/blue bold underline]"
        notebook_name += "[blue bold]:[/blue bold] "
        notebook_name += f"[green]{self.notebook_metadata.notebook_name}[/green]\n"
        notebook_name += "[blue bold]    PATH:[/blue bold] "
        notebook_name += f"[grey50]{self.notebook.path.parent}/"
        notebook_name += f"[bold]{self.notebook.path.name}[bold][/grey50]\n"
        yield notebook_name

        # Statistics panels
        yield "\n[blue bold]STATS[/blue bold]\n"

        # Cells stats
        cells_stats = "\n"
        cells_stats += "[green]Total cells[/green]: "
        cells_stats += f"{self.notebook_stats.number_of_cells}\n"
        cells_stats += "[green]Code cells[/green]: "
        cells_stats += f"{self.notebook_stats.number_of_code_cells}\n"
        cells_stats += "[green]Markdown cells[/green]: "
        cells_stats += f"{self.notebook_stats.number_of_MD_cells}\n"
        cells_stats += "[green]Raw cells[/green]: "
        cells_stats += f"{self.notebook_stats.number_of_raw_cells}\n"

        # Markdown stats
        md_stats = "\n"
        md_stats += "[green]Markdown titles[/green]: "
        md_stats += f"{self.notebook_stats.number_of_md_titles}\n"
        md_stats += "[green]Markdown lines[/green]: "
        md_stats += f"{self.notebook_stats.number_of_md_lines}\n"

        # Modularization stats
        modularization_stats = "\n"
        modularization_stats += "[green]Number of functions[/green]: "
        modularization_stats += f"{self.notebook_stats.number_of_functions}\n"
        modularization_stats += "[green]Number of classes[/green]: "
        modularization_stats += f"{self.notebook_stats.number_of_classes}\n"

        metadata_panels = [
            Panel(cells_stats, title="Cells"),
            Panel(md_stats, title="Markdown usage"),
            Panel(modularization_stats, title="Code modularization"),
        ]
        yield Columns(
            metadata_panels,
            equal=True,
        )

        # Linting results
        if self.has_linting_results:
            yield "\n[blue bold]RESULTS[/blue bold]\n"
            yield self.get_renderable_linting_results()

        yield Rule()



from typing import List, Set

from .config import settings
from .lint import LintDefinition, LintLevel

enabled_cell_level_lints: List[LintDefinition] = []
enabled_notebook_level_lints: List[LintDefinition] = []
enabled_path_level_lints: List[LintDefinition] = []
enabled_project_level_lints: List[LintDefinition] = []


def exclude_lints(
    available_lints: List[LintDefinition], lints_to_exclude: Set[str]
) -> List[LintDefinition]:
    """Filter the list of lint definitions based on the setting ``excluded_lints``."""
    return [lint for lint in available_lints if lint.slug not in lints_to_exclude]


def include_lints(
    available_lints: List[LintDefinition], lints_to_include: Set[str]
) -> List[LintDefinition]:
    """Filter the list of lint definitions based on the setting ``included_lints``."""

    return [lint for lint in available_lints if lint.slug in lints_to_include]


def register_lints(lint_level: LintLevel, lint_defs: List[LintDefinition]) -> None:

    # filter the list of lint definitions based on the settings
    if settings.excluded_lints:
        filtered_lint_defs = exclude_lints(lint_defs, settings.excluded_lints)
    elif settings.included_lints:
        filtered_lint_defs = include_lints(lint_defs, settings.included_lints)
    else:
        filtered_lint_defs = lint_defs

    # Add the filtered list of lints to the global list
    if lint_level == LintLevel.CELL:
        enabled_cell_level_lints.extend(filtered_lint_defs)
    elif lint_level == LintLevel.NOTEBOOK:
        enabled_notebook_level_lints.extend(filtered_lint_defs)
    elif lint_level == LintLevel.PATH:
        enabled_path_level_lints.extend(filtered_lint_defs)
    elif lint_level == LintLevel.PROJECT:
        enabled_project_level_lints.extend(filtered_lint_defs)



import dataclasses
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from pydantic import BaseModel
from rich.columns import Columns
from rich.console import Console, ConsoleOptions, RenderResult, group
from rich.panel import Panel
from rich.rule import Rule

from .lint import PathLevelLint, ProjectLevelLint, RepoLint
from .lint_register import enabled_path_level_lints, enabled_project_level_lints
from .nb_linter import NotebookLinter
from .repository import Repository


class RepoLinterOptions(BaseModel):
    pass


@dataclass
class RepositoryMetadata:
    repository_name: str


@dataclass
class RepositoryStats:
    number_of_notebooks: int


class RepoLinter:
    def __init__(self, repo: Repository) -> None:
        self.repo = repo
        self.options: RepoLinterOptions = RepoLinterOptions()
        self.repository_metadata: RepositoryMetadata = RepositoryMetadata(
            repository_name=repo.path.name or Path.cwd().name
        )
        self.repository_stats: RepositoryStats = RepositoryStats(
            number_of_notebooks=len(repo.notebooks)
        )

        self.lints: List[RepoLint] = []

        self.lints.extend(
            [
                ProjectLevelLint(
                    lint.slug,
                    lint.description,
                    lint.recommendation,
                    lint.linting_function,
                    self.repo,
                )
                for lint in enabled_project_level_lints
            ]
        )

        self.lints.extend(
            [
                PathLevelLint(
                    lint.slug,
                    lint.description,
                    lint.recommendation,
                    lint.linting_function,
                    self.repo,
                )
                for lint in enabled_path_level_lints
            ]
        )

        self.has_linting_results = any([lint.result for lint in self.lints])

        self.notebook_linters = [
            NotebookLinter(notebook) for notebook in self.repo.notebooks
        ]

        self.has_notebook_level_linting_results = any(
            [linter.has_linting_results for linter in self.notebook_linters]
        )

    @group()
    def get_renderable_linting_results(self):
        for lint in self.lints:
            if lint.result:
                yield lint

    def as_dict(self) -> Dict:
        results_dict = {
            "repository_metadata": dataclasses.asdict(self.repository_metadata),
            "repository_stats": dataclasses.asdict(self.repository_stats),
            "lints": [lint.as_dict() for lint in self.lints if lint.result],
            "notebook_level_lints": [
                nb_linter.as_dict() for nb_linter in self.notebook_linters
            ],
        }
        return results_dict

    @group()
    def get_renderable_nblevel_linting_results(self):
        for linter in self.notebook_linters:
            yield linter

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        # Repository name
        repo_name = "\n"
        repo_name += "[blue bold underline]REPOSITORY[/blue bold underline]"
        repo_name += "[blue bold]:[/blue bold] "
        repo_name += f"[green]{self.repository_metadata.repository_name}[/green]\n"
        repo_name += "[blue bold]      PATH:[/blue bold] "
        repo_name += f"[grey50]{self.repo.path.resolve().parent}/"
        repo_name += f"[bold]{self.repo.path.resolve().name}[bold][/grey50]\n"
        yield repo_name

        # Statistics panels

        # Stats
        repo_stats = "\n"
        repo_stats += "[green]Analyzed notebooks[/green]: "
        repo_stats += f"{self.repository_stats.number_of_notebooks}\n"

        metadata_panels = [Panel(repo_stats, title="Stats")]
        yield Columns(metadata_panels, equal=True)
        yield "\n\n"

        # Repo-level linting results
        if self.has_linting_results:
            yield Rule(
                "[turquoise2 bold]REPOSITORY-LEVEL RESULTS[/turquoise2 bold]",
                align="left",
                style="",
            )
            yield "\n"
            yield self.get_renderable_linting_results()
            yield "\n\n\n"

        # Notebook-level linting results
        if self.has_notebook_level_linting_results:
            yield Rule(
                "[turquoise2 bold]NOTEBOOK-LEVEL RESULTS[/turquoise2 bold]",
                align="left",
                style="",
            )
            yield "\n"
            yield self.get_renderable_nblevel_linting_results()
        yield "\n"



""" Entry point of pynblint. Used when running pynblint from the command line. """

import json
import sys
from pathlib import Path
from typing import Union

import typer
from rich.console import Console

from . import loader
from .config import CellRenderingMode, settings
from .exceptions import ExportFormatNotSupportedError
from .nb_linter import NotebookLinter
from .notebook import Notebook
from .repo_linter import RepoLinter
from .repository import GitHubRepository, LocalRepository, Repository

app = typer.Typer()
console = Console(force_terminal=True)


loader.load_core_modules()
loader.load_plugins(settings.plugins)


@app.command()
def main(
    source: str = typer.Argument(..., exists=True),
    from_github: bool = typer.Option(
        None, help="Whether to interpret the source as the URL of a GitHub repository."
    ),
    output_file: Path = typer.Option(
        None,
        "--output",
        "--output-file",
        "-o",
        help="Path of the output file. Based on the extension of the specified filename"
        "Pynblint will chose the output format.\n"
        "Currently, the only supported export format is 'JSON' (extension: `.json`).",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Run pynblint non-interactively by always answering 'yes' "
        "to command-line prompts.",
    ),
    render_full_cells: bool = typer.Option(
        None,
        help="Whether to render full cells or just the first & last line of each cell.",
    ),
    display_cell_index: bool = typer.Option(
        None,
        help="Whether to display cell index \
            (i.e., the zero-based position of the cell within the notebook) \
            above rendered cells.",
    ),
    initial_cells: int = typer.Option(
        None,
        help="The number of cells (from the notebook beginning) that Pynblint should "
        'consider as the "initial cells" of the notebook',
    ),
    final_cells: int = typer.Option(
        None,
        help="The number of cells (from the notebook ending) that Pynblint should "
        'consider as the "final cells" of the notebook',
    ),
    min_md_code_ratio: float = typer.Option(
        None,
        help="The minimum ratio of Markdown cells over code cells. "
        f"Defaults to {settings.min_md_code_ratio}.",
    ),
):

    # Update settings
    if render_full_cells:
        settings.cell_rendering_mode = CellRenderingMode.FULL

    if display_cell_index:
        settings.display_cell_index = True

    if initial_cells:
        settings.initial_cells = initial_cells

    if final_cells:
        settings.final_cells = final_cells

    if min_md_code_ratio:
        settings.min_md_code_ratio = min_md_code_ratio

    # Prevent accidental overwriting of previous output
    if output_file and output_file.is_file() and not yes:
        console.print("[red bold]The specified output file already exists.[/red bold]")
        ans = input("Do you want to overwrite it? (y/n): ")
        if ans.lower() != "y":
            sys.exit()

    # Main procedure
    console.print("\n")
    console.rule("PYNBLINT", characters="*")
    repo: Repository
    linter: Union[NotebookLinter, RepoLinter]

    if from_github:
        # Analyze GitHub repository
        repo = GitHubRepository(source)
        linter = RepoLinter(repo)

    else:
        path: Path = Path(source)

        if path.is_dir():
            # Analyze local uncompressed directory
            repo = LocalRepository(path)
            linter = RepoLinter(repo)

        elif path.suffix == ".ipynb":
            # Analyze standalone notebook
            with open(path) as notebook_file:
                nb = Notebook(Path(notebook_file.name))
                linter = NotebookLinter(nb)

        else:
            # Analyze local compressed directory
            repo = LocalRepository(path)
            linter = RepoLinter(repo)

    if output_file:
        if output_file.suffix == ".json":
            with open(output_file, "w") as f:
                json.dump(linter.as_dict(), f)
        else:
            raise ExportFormatNotSupportedError(
                f"The specified output file extension `{output_file.suffix}` "
                "is not supported yet."
            )

    console.print(linter)


if __name__ == "__main__":
    app()



import os
import tempfile
import zipfile
from abc import ABC
from pathlib import Path
from typing import List, Optional

import git

from .config import settings
from .notebook import Notebook


class Repository(ABC):
    """
    This class stores data about a code repository.
    """

    def __init__(self, path: Path):

        # Repository info
        self.path = path

        # Extracted content
        self.notebooks: List[Notebook] = []  # List of Notebook objects

    def retrieve_notebooks(self):

        # Directories to ignore while traversing the tree
        dirs_ignore = [".ipynb_checkpoints"]

        for root, dirs, files in os.walk(self.path):
            # `dirs[:] = value` modifies dirs in-place
            dirs[:] = [d for d in dirs if d not in dirs_ignore]
            for f in files:
                if f.endswith(".ipynb"):
                    nb = Notebook(Path(root) / Path(f))
                    self.notebooks.append(nb)

    @property
    def is_git_repository(self):
        # Directories to ignore while traversing the tree
        dirs_ignore = [".ipynb_checkpoints"]
        versioned = False
        for _, dirs, _ in os.walk(self.path):
            # `dirs[:] = value` modifies dirs in-place
            dirs[:] = [d for d in dirs if d not in dirs_ignore]
            for d in dirs:
                if d == ".git":
                    versioned = True
        return versioned

    @property
    def large_file_paths(self) -> List[Path]:
        """Return the list of files whose size is above the fixed threshold.

        The threshold size for data files is defined in the settings.

        Returns:
            List[Path]: the list of large files.
        """
        large_files: List[Path] = []
        for dirpath, _, filenames in os.walk(self.path):
            for filename in filenames:
                file_path = Path(dirpath) / filename
                if os.path.getsize(file_path) > settings.max_data_file_size:
                    large_files.append(file_path)
        return large_files


class LocalRepository(Repository):
    """
    This class stores data about a local code repository.
    The `source_path` can point either to a local directory or a zip archive
    """

    def __init__(self, source_path: Path):

        self.source_path = source_path
        tmp_dir: Optional[tempfile.TemporaryDirectory] = None

        # Handle .zip archives
        if self.source_path.suffix == ".zip":

            # Create temp directory
            tmp_dir = tempfile.TemporaryDirectory()
            repo_path: Path = Path(tmp_dir.name)

            # Extract the zip file into the temp folder
            with zipfile.ZipFile(self.source_path, "r") as zip_file:
                zip_file.extractall(repo_path)

        # Handle local folders
        elif self.source_path.is_dir():
            repo_path = self.source_path

        else:
            raise ValueError(
                "The file at the specified path is neither a notebook (.ipynb) "
                "nor a compressed archive."
            )

        super().__init__(repo_path)
        self.retrieve_notebooks()

        # Clean up the temp directory if one was created
        if tmp_dir is not None:
            tmp_dir.cleanup()


class GitHubRepository(Repository):
    """
    This class stores data about a GitHub repository
    """

    def __init__(self, github_url: str):

        self.url = github_url

        # Clone the repo in a temp directory
        tmp_dir = tempfile.TemporaryDirectory()
        git.Repo.clone_from(  # type: ignore
            url=github_url, to_path=tmp_dir.name, depth=1
        )
        super().__init__(Path(tmp_dir.name) / github_url.split("/")[-1])

        # Analyze the repo
        self.retrieve_notebooks()

        # Clean up the temp directory
        tmp_dir.cleanup()



import re
from enum import Enum
from typing import Dict

import rich
from nbformat.notebooknode import NotebookNode
from rich.abc import RichRenderable
from rich.columns import Columns
from rich.padding import Padding
from rich.panel import Panel
from rich.syntax import Syntax

from .config import CellRenderingMode, settings
from .rich_extensions import NotebookMarkdown


class CellType(str, Enum):
    MARKDOWN = "markdown"
    CODE = "code"
    RAW = "raw"
    OTHER = "other"


class Cell(RichRenderable):
    def __init__(self, cell_index: int, cell_dict: NotebookNode) -> None:
        """Pynblint's representation of a notebook cell."""

        self.cell_index: int = cell_index
        self._cell_dict: NotebookNode = cell_dict

        # Cell type
        self.cell_type: CellType
        if self._cell_dict["cell_type"] == "markdown":
            self.cell_type = CellType.MARKDOWN
        elif self._cell_dict["cell_type"] == "code":
            self.cell_type = CellType.CODE
        elif self._cell_dict["cell_type"] == "raw":
            self.cell_type = CellType.RAW
        else:
            self.cell_type = CellType.OTHER

        # Execution count
        if self.cell_type == CellType.CODE:
            self.exec_count: int = self._cell_dict["execution_count"]

        # Cell source and source excerpt (for cell visualization with Rich)
        self.cell_source: str = self._cell_dict["source"]
        self._source_excerpt: str = ""

        if settings.cell_rendering_mode == CellRenderingMode.COMPACT:
            cell_source_array = self.cell_source.split("\n")
            if len(cell_source_array) > 2:
                self._source_excerpt = "\n".join(
                    [cell_source_array[0], "\n[...]\n", cell_source_array[-1]]
                )
            else:
                self._source_excerpt = self.cell_source
        else:
            self._source_excerpt = self.cell_source

    @property
    def empty(self) -> bool:
        """Return ``True`` if the code cell is empty."""
        if self.cell_type == CellType.CODE:
            return (self.exec_count is None) and (len(self.cell_source) == 0)
        else:
            raise Exception("The `empty` property is defined only for code cells.")

    @property
    def non_executed(self) -> bool:
        """Return ``True`` if the code cell is not empty and has not been executed."""
        if self.cell_type == CellType.CODE:
            return (self.exec_count is None) and (len(self.cell_source) > 0)
        else:
            raise Exception(
                "The `non_executed` property is defined only for code cells."
            )

    @property
    def is_heading(self) -> bool:
        """Return ``True`` if the cell is an MD cell containing only MD headings."""
        if self.cell_type == CellType.MARKDOWN:
            pattern = re.compile(r"^\s*#{1,6}\s*[^#\n]*$")
            return all(
                pattern.match(line)
                for line in self.cell_source.splitlines()
                if line and (not line.isspace())
            )
        else:
            return False

    def as_dict(self, source: bool = True) -> Dict:
        cell_dict = {
            "index": self.cell_index,
            "type": str(self.cell_type),
        }
        if self.cell_type == CellType.CODE:
            cell_dict["execution_count"] = self.exec_count
        if source:
            cell_dict["source"] = self._source_excerpt
        return cell_dict

    def __rich__(self) -> Columns:

        if self.cell_type == CellType.CODE:
            counter = self.exec_count or " "
            if settings.display_cell_index:
                panel_title = f"Index: {self.cell_index}"
            else:
                panel_title = None
            rendered_cell = Columns(
                [
                    f"\nIn [{counter}]:",
                    Panel(
                        Syntax(
                            self._source_excerpt, "python", background_color="default"
                        ),
                        width=int(rich.get_console().size[0] * 0.85),
                        title=panel_title,
                    ),
                ]
            )
        else:
            rendered_cell = Columns(
                [
                    "        ",
                    Padding(NotebookMarkdown(self._source_excerpt), (1, 0, 1, 8)),
                ]
            )

        return rendered_cell



from pathlib import Path
from typing import List

import nbconvert
import nbformat
from nbformat.notebooknode import NotebookNode
from rich.abc import RichRenderable
from rich.console import Console, ConsoleOptions, RenderResult

from .cell import Cell, CellType
from .config import settings


class Notebook(RichRenderable):
    """
    This class stores the representations of a notebook
    on which pynblint functions are called
    """

    def __init__(self, path: Path):
        self.path: Path = path

        # Read the notebook with nbformat
        with open(self.path) as f:
            nb_raw = f.read()
        self.nb_dict: NotebookNode = nbformat.reads(nb_raw, as_version=4)

        # Populate the list of Cells
        self.cells: List[Cell] = [
            Cell(cell_index, cell_dict)
            for cell_index, cell_dict in enumerate(self.nb_dict.cells)
        ]

        # Convert the notebook to a Python script
        python_exporter = nbconvert.PythonExporter()
        self.script, _ = python_exporter.from_notebook_node(self.nb_dict)

    @property
    def code_cells(self) -> List[Cell]:
        code_cells = [cell for cell in self.cells if cell.cell_type == CellType.CODE]
        return code_cells

    @property
    def markdown_cells(self) -> List[Cell]:
        md_cells = [cell for cell in self.cells if cell.cell_type == CellType.MARKDOWN]
        return md_cells

    @property
    def initial_cells(self) -> List[Cell]:
        return self.cells[: settings.initial_cells]

    @property
    def final_cells(self) -> List[Cell]:
        return self.cells[-settings.final_cells :]  # noqa: E203

    def __len__(self) -> int:
        return len(self.cells)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        return self.cells



"""Linting functions for repositories containing notebooks"""

from pathlib import Path
from typing import List

from . import lint_register as register
from .lint import LintDefinition, LintLevel
from .repository import Repository

# ============= #
# PROJECT LEVEL #
# ============= #


def repository_not_versioned(repo: Repository) -> bool:
    """Check the absence of the ``.git`` folder."""

    return not repo.is_git_repository


def dependencies_unmanaged(repo: Repository) -> bool:
    """Check the absence of configuration files for dependency management tools.

    All configuration files are searched in the root of the repository.
    """

    paths = [
        repo.path / "requirements.txt",
        repo.path / "pyproject.toml",
        repo.path / "environment.yml",
        repo.path / "setup.py",
        repo.path / "Pipfile",
    ]

    return not any(map(lambda x: x.exists(), paths))


# ========== #
# PATH LEVEL #
# ========== #


def duplicate_notebook_filename(repo: Repository) -> List[Path]:
    """Check the existence of notebooks with the same filename within a repository"""

    nb_filenames = []
    duplicate_filanames = []
    paths = []
    for notebook in repo.notebooks:
        filename = notebook.path.name
        if filename in nb_filenames:
            duplicate_filanames.append(filename)
        else:
            nb_filenames.append(filename)
    for filename in duplicate_filanames:
        for notebook in repo.notebooks:
            if notebook.path.name == filename:
                paths.append(notebook.path)
    return paths


def unversioned_large_data_files(repo: Repository) -> List[Path]:
    """Check the presence of unversioned large data files.

    Check wether large data files are versioned with a data version control system.

    Currently, the only data VCS that Pynblint detects is DVC (https://dvc.org).
    Alternative solutions will be added soon.

    Args:
        repo (Repository): The repository to be analyzed.

    Returns:
        List[Path]: the list of large data files that should be put under
        version control.
    """

    large_files = repo.large_file_paths
    if large_files and not Path(repo.path, ".dvc").is_dir():
        return large_files
    else:
        return []


# ================= #
# LINT REGISTRATION #
# ================= #


project_level_lints: List[LintDefinition] = [
    LintDefinition(
        slug="repository-not-versioned",
        description="This repository is not version controlled.",
        recommendation="Put the repository under version control using a VCS like git.",
        linting_function=repository_not_versioned,
    ),
    LintDefinition(
        slug="dependencies-unmanaged",
        description="No dependency-management tools appear to be used in this project.",
        recommendation="If you are using `pip`, declare your dependencies in a "
        "`requirements.txt` file. You can do so by running the following command: "
        "`pip freeze > requirements.txt`.",
        linting_function=dependencies_unmanaged,
    ),
]

path_level_lints: List[LintDefinition] = [
    LintDefinition(
        slug="duplicate-notebook-filename",
        description="Two or more notebooks with the same filename exist in this "
        "repository.",
        recommendation="Use different filenames to make notebooks easy to recognize; "
        "possibly stick to a naming convention.",
        linting_function=duplicate_notebook_filename,
    ),
    LintDefinition(
        slug="large-data-file-not-versioned",
        description="Your repository contains one or more data files that are not "
        "versioned.",
        recommendation="Use DVC (https://dvc.org) to put your data under "
        "version control;\nmake your data available by pushing it to one of the "
        "supported remotes.",
        linting_function=unversioned_large_data_files,
    ),
]


def initialize() -> None:
    register.register_lints(LintLevel.PROJECT, project_level_lints)
    register.register_lints(LintLevel.PATH, path_level_lints)



from enum import Enum
from typing import List, Optional, Set

from pydantic import BaseSettings


class CellRenderingMode(str, Enum):
    FULL = "full"
    COMPACT = "compact"


class Settings(BaseSettings):

    plugins: List[str] = []
    included_lints: Optional[Set[str]] = None
    excluded_lints: Optional[Set[str]] = None
    cell_rendering_mode: CellRenderingMode = CellRenderingMode.COMPACT
    result_details_indentation: int = 5
    display_cell_index: bool = False
    filename_max_length: int = 0
    max_cells_in_notebook: int = 50
    max_lines_in_code_cell: int = 30
    initial_cells: int = 3
    final_cells: int = 3
    min_md_code_ratio: float = 0.3
    max_data_file_size: int = 10 * 1000000  # 10MB

    # TODO: custom validation: included_lints OR excluded lints must be None
    #       I.e., something like:
    #
    # if (excluded_lints is not None) and (included_lints is not None):
    #     raise ValueError(
    #         "The arguments `excluded_lints` and `included_lints` cannot be used \
    #             at the same time. \
    #         Please, choose whether you need to specify the lints to be exluded \
    #             or those to be included."
    #     )

    class Config:
        env_file = ".pynblint"
        env_file_encoding = "utf-8"


settings = Settings()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pynblint']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.26,<4.0.0',
 'ipython<8',
 'nbconvert>=6.4.0,<7.0.0',
 'nbformat>=5.1.3,<6.0.0',
 'pydantic[dotenv]>=1.9.0,<2.0.0',
 'rich>=11.1.0,<12.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['pynblint = pynblint.main:app']}

setup_kwargs = {
    'name': 'pynblint',
    'version': '0.1.3',
    'description': 'A linter for Jupyter notebooks written in Python.',
    'long_description': '# Pynblint\n\n[![CI](https://github.com/collab-uniba/pynblint/actions/workflows/CI.yml/badge.svg)](https://github.com/collab-uniba/pynblint/actions/workflows/CI.yml)\n[![Documentation Status](https://readthedocs.org/projects/pynblint/badge/?version=latest)](https://pynblint.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/collab-uniba/pynblint/branch/master/graph/badge.svg?token=CSX10BJ1CU)](https://codecov.io/gh/collab-uniba/pynblint)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![code style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nMany professional data scientists use Jupyter Notebook to accomplish their daily tasks, from preliminary data exploration to model prototyping. Notebooks\' interactivity is particularly convenient for data-centric programming; moreover, their self-documenting nature greatly simplifies and enhances the communication of analytical results.\n\nHowever, Jupyter Notebook has often been criticized for offering scarce native support for Software Engineering best practices and inducing bad programming habits. To really benefit from computational notebooks, practitioners need to be aware of their common pitfalls and learn how to avoid them.\n\nIn our paper ["Eliciting Best Practices for Collaboration with Computational Notebooks" [1]](https://arxiv.org/abs/2202.07233), we introduced a catalog of validated best practices for the collaborative use of notebooks in professional contexts.\n\nTo raise awareness of these best practices and promote their adoption, we have created Pynblint, a static analysis tool for Jupyter notebooks written in Python. Pynblint can be operated as a standalone CLI application or as part of a CI/CD pipeline. It reveals potential defects of Jupyter notebooks found in software repositories and recommends corrective actions.\n\nThe core linting rules that power Pynblint have been derived as operationalizations of the validated best practices from our catalog. Nonetheless, the tool is designed to be easily customized and extended with new rules.\n\n## Catalog of best practices\n\n- Use version control\n- Manage project dependencies\n- Use self-contained environments\n- Put imports at the beginning\n- Ensure re-executability (re-run notebooks top to bottom)\n- Modularize your code\n- Test your code\n- Name your notebooks consistently\n- Stick to coding standards\n- Use relative paths\n- Document your analysis\n- Leverage Markdown headings to structure your notebook\n- Keep your notebook clean\n- Keep your notebook concise\n- Distinguish production and development artifacts\n- Make your notebooks available\n- Make your data available\n\n## Installation\n\nTo use Pynblint, clone this repository and install it with [Poetry](https://python-poetry.org):\n\n```bash\npoetry install --no-dev\n```\n\nTo install Pynblint for development purposes, simply omit the `--no-dev` option:\n\n```bash\npoetry install\n```\n\nAt present, we are finalizing the first version of Pynblint (v0.1.0).\nWhen released, it will become available as a Python package on PyPI and installable via `pip`.\n\n\n## Usage\n\nOnce installed, Pynblint can be used to analyze:\n\n- a single notebook:\n\n    ```bash\n    pynblint path/to/the/notebook.ipynb\n    ```\n- the set of notebooks found in the current working directory:\n\n    ```bash\n    pynblint .\n    ```\n\n- the set of notebooks found in the directory located at the specified path:\n\n    ```bash\n    pynblint path/to/the/project/dir/\n    ```\n\n- the set of notebooks found in a compressed `.zip` archive:\n\n    ```bash\n    pynblint path/to/the/compressed/archive.zip\n    ```\n\n- the set of notebooks found in a _public_ GitHub repository (support for private repositories is on our roadmap 🙂):\n\n    ```bash\n    pynblint --from-github https://github.com/collab-uniba/pynblint\n    ```\n\nFor further information on the available options, please read Pynblint\'s CLI manual:\n\n```bash\npynblint --help\n```\n\n## References\n\nLuigi Quaranta, Fabio Calefato, and Filippo Lanubile. 2022. [Eliciting Best Practices for Collaboration with Computational Notebooks.](https://arxiv.org/abs/2202.07233) *Proc. ACM Hum.-Comput. Interact.* 6, CSCW1, Article 87 (April 2022), 41 pages. <https://doi.org/10.1145/3512934>\n',
    'author': 'Luigi Quaranta',
    'author_email': 'luigi.quaranta@uniba.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pynblint.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Linting modules loader."""

import importlib
from typing import List

from . import nb_linting, repo_linting


def load_core_modules() -> None:
    nb_linting.initialize()
    repo_linting.initialize()


class PluginInterface:
    """A plugin has a single function called initialize."""

    @staticmethod
    def initialize() -> None:
        """Initialize the plugin."""


def import_module(name: str) -> PluginInterface:
    return importlib.import_module(name)  # type: ignore


def load_plugins(plugins: List[str]) -> None:
    """Load the plugins defined in the plugins list."""

    for plugin_name in plugins:
        plugin = import_module(plugin_name)
        plugin.initialize()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from rich.console import Console, ConsoleOptions, RenderResult, group
from rich.padding import Padding

from .config import settings
from .notebook import Cell, Notebook
from .repository import Repository


class LintLevel(str, Enum):
    CELL = "cell"
    NOTEBOOK = "notebook"
    PATH = "path"
    PROJECT = "project"


@dataclass
class LintDefinition:
    slug: str
    description: str
    recommendation: str
    linting_function: Callable
    show_details: bool = True


# ============== #
# NOTEBOOK LINTS #
# ============== #


class NotebookLint(ABC):
    def __init__(self, slug: str, description: str, recommendation: str) -> None:
        super().__init__()
        self.slug: str = slug
        self.description: str = description
        self.recommendation: str = recommendation
        self.result: Optional[Union[bool, List[Cell]]] = None

    @abstractmethod
    def lint(self, notebook: Notebook) -> Union[bool, List[Cell]]:
        pass

    def as_dict(self, description: bool = True, recommendation: bool = True) -> Dict:
        lint_dict = {
            "slug": self.slug,
        }
        if description:
            lint_dict["description"] = self.description
        if recommendation:
            lint_dict["recommendation"] = self.recommendation
        return lint_dict

    @abstractmethod
    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        pass


class NotebookLevelLint(NotebookLint):
    def __init__(
        self,
        slug: str,
        description: str,
        recommendation: str,
        linting_function: Callable[[Notebook], bool],
        notebook: Notebook,
    ) -> None:
        super().__init__(slug, description, recommendation)
        self.linting_function: Callable[[Notebook], bool] = linting_function
        self.result: bool = self.lint(notebook)

    def lint(self, notebook: Notebook) -> bool:
        return self.linting_function(notebook)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield f"[orange3 bold]({self.slug})[/orange3 bold]: {self.description}"
        if len(self.recommendation):
            yield Padding(
                f"[italic]Recommendation[/italic]: {self.recommendation}",
                (0, 0, 0, settings.result_details_indentation),
            )
        yield "\n"


class CellLevelLint(NotebookLint):
    def __init__(
        self,
        slug: str,
        description: str,
        recommendation: str,
        linting_function: Callable[[Notebook], List[Cell]],
        notebook: Notebook,
        show_details: bool = True,
    ) -> None:
        super().__init__(slug, description, recommendation)
        self.linting_function: Callable[[Notebook], List[Cell]] = linting_function
        self.show_details = show_details
        self.result: List[Cell] = self.lint(notebook)

    def lint(self, notebook: Notebook) -> List[Cell]:
        """
        Execute the cell-level lint represented by the class.

        Returns:
            The index list of affected cells in case of a positive result,
            an empty list otherwise.
        """
        return self.linting_function(notebook)

    def as_dict(self, description: bool = True, recommendation: bool = True) -> Dict:
        lint_dict = super().as_dict(description, recommendation)
        lint_dict["cells"] = [cell.as_dict(source=False) for cell in self.result]
        return lint_dict

    @group()
    def get_renderable_affected_cells(self):
        for cell in self.result:
            yield Padding(cell, (0, 0, 0, settings.result_details_indentation))

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield f"[orange3 bold]({self.slug})[/orange3 bold]: {self.description}"
        if len(self.recommendation):
            yield Padding(
                f"[italic]Recommendation[/italic]: {self.recommendation}",
                (0, 0, 0, settings.result_details_indentation),
            )

        yield Padding(
            "[italic]Affected cells[/italic]: [grey50]indexes[/grey50]"
            + str([cell.cell_index for cell in self.result]),
            (0, 0, 0, settings.result_details_indentation),
        )
        if self.show_details:
            yield self.get_renderable_affected_cells()
        yield "\n"


# ================ #
# REPOSITORY LINTS #
# ================ #


class RepoLint(ABC):
    def __init__(self, slug: str, description: str, recommendation: str) -> None:
        super().__init__()
        self.slug: str = slug
        self.description: str = description
        self.recommendation: str = recommendation
        self.result: Optional[Union[bool, List[Path]]] = None

    @abstractmethod
    def lint(self, repository: Repository) -> Union[bool, List[Path]]:
        pass

    def as_dict(self, description: bool = True, recommendation: bool = True) -> Dict:
        lint_dict = {
            "slug": self.slug,
        }
        if description:
            lint_dict["description"] = self.description
        if recommendation:
            lint_dict["recommendation"] = self.recommendation
        return lint_dict

    @abstractmethod
    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        pass


class ProjectLevelLint(RepoLint):
    def __init__(
        self,
        slug: str,
        description: str,
        recommendation: str,
        linting_function: Callable[[Repository], bool],
        repository: Repository,
    ) -> None:
        super().__init__(slug, description, recommendation)
        self.linting_function: Callable[[Repository], bool] = linting_function
        self.result: bool = self.lint(repository)

    def lint(self, repository: Repository) -> bool:
        return self.linting_function(repository)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield f"[blue bold]({self.slug})[/blue bold]: {self.description}"
        if len(self.recommendation):
            yield Padding(
                f"[italic]Recommendation[/italic]: {self.recommendation}",
                (0, 0, 0, settings.result_details_indentation),
            )
        yield "\n"


class PathLevelLint(RepoLint):
    def __init__(
        self,
        slug: str,
        description: str,
        recommendation: str,
        linting_function: Callable[[Repository], List[Path]],
        repository: Repository,
    ) -> None:
        super().__init__(slug, description, recommendation)
        self.linting_function: Callable[[Repository], List[Path]] = linting_function
        self.result: List[Path] = self.lint(repository)

    def lint(self, repository: Repository) -> List[Path]:
        return self.linting_function(repository)

    def as_dict(self, description: bool = True, recommendation: bool = True) -> Dict:
        lint_dict = super().as_dict(description, recommendation)
        lint_dict["paths"] = [str(path) for path in self.result]
        return lint_dict

    @group()
    def get_renderable_affected_cells(self):
        for path in self.result:
            yield Padding(
                "  •  [yellow]" + str(path) + "[/yellow]",
                (0, 0, 0, settings.result_details_indentation),
            )

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield f"[blue bold]({self.slug})[/blue bold]: {self.description}"
        if len(self.recommendation):
            yield Padding(
                f"[italic]Recommendation[/italic]: {self.recommendation}",
                (0, 0, 0, settings.result_details_indentation),
            )

        yield Padding(
            "[italic]Affected paths[/italic]:",
            (0, 0, 0, settings.result_details_indentation),
        )
        yield self.get_renderable_affected_cells()
